# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
"""
Version information.
"""

# the major version
pyosmium_major = '4.0'
# current release (Pip version)
pyosmium_release = '4.0.2'

# libosmium version shipped with the Pip release
libosmium_version = '2.20.0'
# protozero version shipped with the Pip release
protozero_version = '1.7.1'
# pybind11 version shipped with the Pip release
pybind11_version = '2.13.6'
